/*
 C++ implementation of paper written for ACM MM 2011 on Video Shot 
 Classification based on cinematographic constratints. The idea is to extend
 the experiments for ALADDIN shots. 
 The input data we are going to work with is small 2-5 seconds clips from 
 action concepts. It can also be 5-10 secs of fixed length shots (depends on 
 frame rate). The homography between the frames in the clips are to be 
 transformed into different lie space representation. 

 input: videofile.mp4
 output: homographies.txt (8 x n matrix)
       : alphas.txt (8 x n matrix) where n is the number of pairs of frames 
         between which homography is computed.

 Here are the algorithmic steps:
 - Read in a video
 - Compute frame to frame homography
 - Obtain Lie subspace representation of the homography coefficients
 - construct time series for all Lie-subspace coefficients
 - compute chaotic invariant/statistical features from coefficients
 
 The next steps - outside the scope of this program is to : 
 - Generate final representation in SVM classification format
 - Perform one against all classification
   
 @author  Subhabrata Bhattacharya
 @email   subh@cs.ucf.edu
 @date    08/11/2012
 @version 1.0
*/

#include <opencv2/calib3d/calib3d.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/features2d/features2d.hpp>
#include <opencv2/legacy/legacy.hpp>

#include <iostream>
#include <vector>
#include <stdio.h>

#define MAX_PATH_LEN 1024
#define H_NUM_MATCHES 5
#define H_ERROR_THRESHOLD 256
#define EXTN_LIECOEFF       ".lie" 
#define EXTN_STATFEAT       ".stat" 


using namespace cv;
using namespace std;

struct params{
  /* Processed output to be used as input by other modules */
  VideoCapture video_src;
  int frame_width;
  int frame_height;
  int video_fps;
  int num_frames; 
  int window_size;

  char output_suffix[MAX_PATH_LEN];
  char homography_file[MAX_PATH_LEN];
  char liespace_rep_file[MAX_PATH_LEN];
  char ts_feat_file[MAX_PATH_LEN];
  char stat_feat_file[MAX_PATH_LEN];
};
typedef struct params P_PARAMS;


namespace {
void get_homography_sequence (VideoCapture capture, vector<Mat>& vec_H);
Mat compute_Liecoeffs (const Mat& H);
void construct_timeseries (const vector<Mat>& vec_L, vector <Mat>& vec_LTS);

double entropy (const Mat& data);
int zero_crossing (const Mat& data);
double corr(const Mat& x, const Mat &y);
Mat compute_stat_feat(const vector<Mat>& vec_LTS);

/******************************************************************/
/* BEGIN Program specific Utility,I/O routines for debug purpose  */
/******************************************************************/
void print_matrix (const Mat& x, const char* marker) {
  fprintf(stderr, "%dx%d matrix: %s\n", x.rows, x.cols, marker);
  for (int icnt =0; icnt < x.rows; icnt++){
    for (int jcnt =0; jcnt < x.cols; jcnt++)
      fprintf(stderr, "%2.5lf ", x.at<double>(icnt, jcnt));
    fprintf(stderr, "\n");
  }
}
void save_matrix (const Mat& x, const char* filename) {
  FILE *fp;
  fprintf(stderr, "%dx%d matrix\n", x.rows, x.cols);
  fp = fopen (filename, "w+"); 
  for (int icnt =0; icnt < x.rows; icnt++){
    for (int jcnt =0; jcnt < x.cols; jcnt++)
      fprintf(fp, "%2.5lf ", x.at<double>(icnt, jcnt));
    fprintf(stderr, "\n");
  }
  fclose(fp);
}

void print_vector_matrix (vector<Mat>& x, const char* marker) {
  while (!x.empty()) {
    print_matrix (x.back(), marker);
    x.pop_back();
  }
}
void save_vectors (const vector<Mat>& x, char *fname){
  FILE *fp;
  int e_count;
  fp = fopen (fname, "w+");
  for (int v_count=0; v_count < x.size(); v_count++) {
    for (e_count =0; e_count < x[v_count].cols-1; e_count++)
      fprintf(fp, "%2.5lf ", x[v_count].at<double>(0, e_count));
    /* Print last element in the vector - to avoid trailing space */
    fprintf(fp, "%2.5lf\n", x[v_count].at<double>(0, e_count));
  }
  fclose(fp);
}

void usage(char** argv) {
  cout << "Usage: " << argv[0] << "\n"
          "         -i /video/file/path\n" 
          "         -w window size\n"
          "         -s output/suffix/file/path\n"
       << endl;
}

/* Parse command line option */
int parse_options (int argc, char **argv, P_PARAMS &p) {
  int i;
  
  if( argc == 1) {
    usage (argv);
    return 1;
  }
  for (i = 1; i < argc; i++ ){
    /* Input video file */
    if (!strcmp (argv[i], "-i")) {
      string str = string(argv[++i]);
      p.video_src.open(str);
      if (!p.video_src.isOpened()){
        cerr << "ERR: Failed to open a video file: \n" << endl;
        usage (argv);
        return 1;
      }
      /* Perform populating the structure with video information */
      p.frame_width = p.video_src.get(CV_CAP_PROP_FRAME_WIDTH);
      p.frame_height = p.video_src.get(CV_CAP_PROP_FRAME_HEIGHT);
      p.video_fps = p.video_src.get(CV_CAP_PROP_FPS);
      p.num_frames = p.video_src.get(CV_CAP_PROP_FRAME_COUNT);
    }
    
    /* window size to chose from (default 1) */
    if (!strcmp (argv[i], "-w")) {
      p.window_size = strtod (argv[++i], NULL);
    }
    
    /* This is used as a base name, different extensions are added for 
    different output files */ 
    if (!strcmp (argv[i], "-s")) {
      strcpy (p.output_suffix, argv[++i]);
      strcpy (p.liespace_rep_file, p.output_suffix);
      strcat(p.liespace_rep_file, EXTN_LIECOEFF);

      strcpy (p.stat_feat_file, p.output_suffix);
      strcat(p.stat_feat_file, EXTN_STATFEAT);
    }
  }
  fprintf (stdout, "INF:Using \n");
       
  fprintf (stdout, "  Video resolution        : %d x %d\n"
                   "  Video Frames per second : %d\n"
                   "  Number of frames        : %d\n"
                   "  Homography Window Size  : %d\n"
                   "  Output Lie Space Rep    : %s\n"
                   "  Output Stat. Feat. file : %s\n",
      p.frame_width, p.frame_height, p.video_fps, p.num_frames, p.window_size, 
      p.liespace_rep_file, p.stat_feat_file);
  return 0;
}
/******************************************************************/
/* END Program specific Utility,I/O routines for debug purpose  */
/******************************************************************/


/* Core function responsible for video processing 
 Control Flow:
   For each pair of frames 
   Compute SIFT/SURF features
   Get corresponding x,y co-ordinates for homography computation
   Compute homography
   Assign  current features to previous frame features (saves re-computation)
   Store homographies in a vector
   Compute Lie-subspace coefficients 
   Store coefficients in a vector
   Construct time series of Lie-coefficients
   Compute Statistical features
   Store Features in a vector
   Output data
*/
int process(P_PARAMS &p)
{
  vector<Mat> vec_H, vec_L, vec_LTS;
  Mat stat_feat;
  fprintf(stderr, "INFO: Computing Homographies.\n");
  get_homography_sequence (p.video_src, vec_H);
  
  fprintf(stderr, "INFO: Computing Lie Coefficients.\n");
  for (int i=0;i<vec_H.size(); i++) {
    Mat L = compute_Liecoeffs (vec_H[i]);
    vec_L.push_back(L);
  }

  /* Output timeseries of Lie coefficients */
  construct_timeseries (vec_L, vec_LTS);
  /* save vectors in output files */
  save_vectors (vec_L, p.liespace_rep_file);

  /* Computing statistical features */
  fprintf(stderr, "INFO: Computing Statisitical Features.\n");
  stat_feat = compute_stat_feat(vec_LTS);
  save_matrix (stat_feat, p.stat_feat_file);

  return 0;
}

/******************************************************************/
/* BEGIN Program specific Core routines                           */
/******************************************************************/

/* 
  Given source and destination keypoints from two subsequent images, 
  find corresponding points using standard descriptor matching 
  using correspondence, compute homography 
  return homography matrix elements in the form of 1x8 vector.
  Generate a sequence of subsequent Homography elements
*/
/* Helper routines for Homography computation */
void drawMatchesRelative(const vector<KeyPoint>& train, const vector<KeyPoint>& query,
  std::vector<cv::DMatch>& matches, Mat& img, const vector<unsigned char>& mask = vector<
  unsigned char> ()) {
  for (int i = 0; i < (int)matches.size(); i++){
    if (mask.empty() || mask[i]){
      Point2f pt_new = query[matches[i].queryIdx].pt;
      Point2f pt_old = train[matches[i].trainIdx].pt;
      Point2f dist = pt_new - pt_old;

      cv::line(img, pt_new, pt_old, Scalar(125, 255, 125), 1);
      cv::circle(img, pt_new, 2, Scalar(255, 0, 125), 1);

    }
  }
}
//Takes a descriptor and turns it into an xy point
void keypoints2points(const vector<KeyPoint>& in, vector<Point2f>& out){
  out.clear();
  out.reserve(in.size());
  for (size_t i = 0; i < in.size(); ++i)
    out.push_back(in[i].pt);
}

//Takes an xy point and appends that to a keypoint structure
void points2keypoints(const vector<Point2f>& in, vector<KeyPoint>& out){
  out.clear();
  out.reserve(in.size());
  for (size_t i = 0; i < in.size(); ++i)
    out.push_back(KeyPoint(in[i], 1));
}

//Uses computed homography H to warp original input points to new planar position
void warpKeypoints(const Mat& H, const vector<KeyPoint>& in, vector<KeyPoint>& out){
  vector<Point2f> pts;
  keypoints2points(in, pts);
  vector<Point2f> pts_w(pts.size());
  Mat m_pts_w(pts_w);
  perspectiveTransform(Mat(pts), m_pts_w, H);
  points2keypoints(pts_w, out);
}

//Converts matching indices to xy points
void matches2points(const vector<KeyPoint>& train, const vector<KeyPoint>& query,
  const std::vector<cv::DMatch>& matches, std::vector<cv::Point2f>& pts_train,
  std::vector<Point2f>& pts_query) {

  pts_train.clear();
  pts_query.clear();
  pts_train.reserve(matches.size());
  pts_query.reserve(matches.size());

  size_t i = 0;
  for (; i < matches.size(); i++) {
    const DMatch & dmatch = matches[i];
    pts_query.push_back(query[dmatch.queryIdx].pt);
    pts_train.push_back(train[dmatch.trainIdx].pt);
  }
}

void resetH(Mat&H){
  H = Mat::eye(3, 3, CV_64F);
}

void get_homography_sequence (VideoCapture capture, vector<Mat>& vec_H) {
  BriefDescriptorExtractor brief(32);
  Mat frame;

  vector<DMatch> matches;
#if CV_SSSE3
    BruteForceMatcher< HammingSeg<30,4> > desc_matcher;
#else
    BruteForceMatcher<Hamming> desc_matcher;
#endif

  vector<Point2f> train_pts, query_pts;
  vector<KeyPoint> train_kpts, query_kpts;
  vector<unsigned char> match_mask;

  Mat gray;

  Mat train_desc, query_desc;
  const int DESIRED_FTRS = 500;
  GridAdaptedFeatureDetector detector(new FastFeatureDetector(10, true), DESIRED_FTRS, 4, 4);
  Mat H;
  Mat H_prev = Mat::eye(3, 3, CV_64F);
  for (;;) {
    capture >> frame;
    if (frame.empty())
      break;

    cvtColor(frame, gray, CV_RGB2GRAY);
    //Find interest points
    detector.detect(gray, query_kpts); 
    //Compute brief descriptors at each keypoint location
    brief.compute(gray, query_kpts, query_desc); 

    if (!train_kpts.empty()){
      vector<KeyPoint> test_kpts;
      warpKeypoints(H_prev.inv(), query_kpts, test_kpts);

      Mat mask = windowedMatchingMask(test_kpts, train_kpts, 25, 25);
      desc_matcher.match(query_desc, train_desc, matches, mask);

      drawKeypoints(frame, test_kpts, frame, Scalar(255, 0, 0), DrawMatchesFlags::DRAW_OVER_OUTIMG);

      matches2points(train_kpts, query_kpts, matches, train_pts, query_pts);

      if (matches.size() > 5){
        H = findHomography(train_pts, query_pts, RANSAC, 4, match_mask);
        if (countNonZero(Mat(match_mask)) > 15) {
          H_prev = H;
        } else {
          resetH(H_prev);
          //drawMatchesRelative(train_kpts, query_kpts, matches, frame, match_mask);
        }
      } else 
        resetH(H_prev);
    } else {
      H_prev = Mat::eye(3, 3, CV_64F);
      Mat out;
      //drawKeypoints(gray, query_kpts, out);
      frame = out;
  
    }
    vec_H.push_back(H_prev.reshape(0,9).t()); 
    //imshow("frame", frame);
    //waitKey(2);
    train_kpts = query_kpts;
    query_desc.copyTo(train_desc);
  }
}

/* 
 Compute Lie space coefficients from homographies: We are using second order 
 approximation of M = log(H) = (H-I) + (-1/2) (H-I)'(H-I)
*/
Mat compute_Liecoeffs (const Mat& h) {
  
  /* MATLAB code
  lb(:,:,1) =   [0  0 1 ;0  0 0; 0 0 0]; %G1 -> (1 ;0) 
  lb(:,:,2) =   [0  0 0 ;0  0 1; 0 0 0]; %G2 -> (0 ;1)
  lb(:,:,3) = 1*[0 -1 0 ;1  0 0; 0 0 0]; %G3 -> (-y;x)
  lb(:,:,4) = 1*[1  0 0 ;0  1 0; 0 0 0]; %G4 -> (x ;y)size(H_data,1)
  lb(:,:,5) = 1*[1  0 0 ;0 -1 0; 0 0 0]; %G5 -> (x ;-y)
  lb(:,:,6) = 1*[0  1 0 ;1  0 0; 0 0 0]; %G6 -> (y ;x)
  lb(:,:,7) = 1*[0  0 0 ;0  0 0; 1 0 0]; %G7 ->
  lb(:,:,8) = 1*[0  0 0 ;0  0 0; 0 1 0]; %G8 ->
  lb = reshape (lb,[], size(lb,3));
  
  k = size(H_data,1);
  alpha_data = zeros(k, 8);
  for i=1:k
    h = reshape(H_data(i,:), 3,3) - eye(3);
    h = reshape(h, 9, []);
    alpha_data(i,:) = h'*lb;  
  end
  */ 
  double data[] = { 
     0, 0, 0, 1, 1, 0, 0, 0,
     0, 0, 1, 0, 0, 1, 0, 0,
     0, 0, 0, 0, 0, 0, 1, 0,
     0, 0,-1, 0, 0, 1, 0, 0,
     0, 0, 0, 1,-1, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 1,
     1, 0, 0, 0, 0, 0, 0, 0,
     0, 1, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0};
  Mat lb = Mat(9, 8, h.type(), data).clone();
  Mat L = Mat::zeros(1, 8, h.type());
  Mat I = Mat::eye(3, 3, h.type());
  Mat res;
  /* 1st order approximation */
  /* res = h.reshape (0, 3) - I;*/
  /* 2nd order approximation */
  res = h.reshape(0, 3)-I + -(1/2)*(h.reshape(0,3)-I).t()*(h.reshape(0,3)-I);
  L = res.reshape(0,9).t() *lb;
  
  res.release();
  lb.release(); 
  return L; 
}

/* Generate Time series of Lie coefficients - This would lead to 8 N-dimensional 
vectors (CvMat of 1 row) where N is the total number of frames in the video. */
void construct_timeseries (const vector<Mat>& vec_L, vector <Mat>& vec_LTS){
  int n = vec_L.size();
  int dim = vec_L.back().cols;
  /* column matrix to store time series */
  Mat x = Mat::zeros(n, 1, vec_L.back().type());
  /* reverse iteration to make sure that the first coefficient time series 
  comes as first element in the vector vec_LTS() */
  for (int j=0; j<dim; j++){ 
    for (int i=0; i<n; i++)
      x.at<double>(i,0)= vec_L[i].at<double>(0, j);
    /* put matrix in vector - the transpose operation copies the data */
    vec_LTS.push_back(x.t());  
  }
  x.release();
}
/* Compute statistical features from time series data. Following are the 
statistical features that are computed. 
For each dimension: 
  mean
  variance
  min
  max
  range
  mean crossing rate [length(crossing(sig-mean(sig)))/length(sig)]
  average root mean square
  mean of skew [mean(((sig-mean(sig)).^3)./(std(sig)^3))]
  variance of skew [mean(((sig-mean(sig)).^3)./(std(sig)^3))]
  entropy according to Shannon entropy formula
  mean of kurtosis [mean((sig-mean(sig)).^4)/(std(sig)^4);]
  variance of kurtosis [mean((sig-mean(sig)).^4)/(std(sig)^4);]

For each dimension pair:
  correlation (7+6+5+4+3+2+1)=28

Sum of signals
Squared sum of signals
Total: 11x8 + 28 + 2 = 118 Dimensional feature vector
*/
Mat compute_stat_feat(const vector<Mat>& vec_LTS){
  int j = 0, sig_length;
  Scalar mu, sigma, s_mu, s_sigma, k_mu, k_sigma;
  MatND mask;
  double min_val, max_val; 
  Point min_loc, max_loc;
  Mat ms, res;
  int dim = vec_LTS.size();
  Mat stat_feat = Mat::zeros (1, 118, vec_LTS.back().type());
  
  /* For independent features */
  for (int i=0; i<dim; i++) {
    /* mean and variance */
    meanStdDev (vec_LTS[i], mu, sigma, mask);
    stat_feat.at<double>(0, j++) = mu.val[0];
    stat_feat.at<double>(0, j++) = pow(sigma.val[0],2);
    /* min, max, range */
    minMaxLoc (vec_LTS[i], &min_val, &max_val, &min_loc, &max_loc, mask);
    stat_feat.at<double>(0, j++) = min_val;
    stat_feat.at<double>(0, j++) = max_val;
    stat_feat.at<double>(0, j++) = max_val - min_val;
    /* Normalized crossing rate */
    Mat ms = vec_LTS[i] - mu.val[0];
    sig_length = vec_LTS[i].cols;
    stat_feat.at<double>(0, j++) = (double)zero_crossing(ms)/sig_length;
    /* Normalized RMS */
    stat_feat.at<double>(0, j++) = (double)sqrt(pow(norm(vec_LTS[i], CV_L2),2)/sig_length);
    /* Computing skew */
    pow(ms, 3, res);
    res = res / pow((double)sigma.val[0], 3);
    meanStdDev (res, s_mu, s_sigma, mask);
    stat_feat.at<double>(0, j++) = s_mu.val[0];
    stat_feat.at<double>(0, j++) = s_sigma.val[0];
    /* Entropy */
    //stat_feat.at<double>(0, j++) = entropy(vec_LTS[i]);
    /* Kurtosis */
    pow(ms, 4, res);
    res = res / pow(sigma.val[0], 4);
    meanStdDev (res, k_mu, k_sigma, mask);
    stat_feat.at<double>(0, j++) = k_mu.val[0];
    stat_feat.at<double>(0, j++) = k_sigma.val[0];
  }
  /* For correlation features */
  Mat flag = Mat::zeros(dim, dim, CV_32FC1);
 
  for (int i=0; i < dim; i++)
    for (int k=0; k < dim; k++) 
      if (i!=k &&  flag.at<float>(i,k)==0 && flag.at<float>(k,i)==0){
        stat_feat.at<double>(0, j++) = corr(vec_LTS[i],  vec_LTS[k]);
        flag.at<float>(i,k)=1;
        flag.at<float>(k,i)=1;
      }
  /* Finally the sum of the signals */
  //sum(abs(sig1)+abs(sig2)+abs(sig3))/size(X,1);
  //sum(sqrt(sig1.^2 + sig2.^2 + sig3.^2))/size(X,1);
  Mat sum_m = Mat::zeros(vec_LTS[0].rows, vec_LTS[0].cols, vec_LTS[0].type());
  Mat sum_sq = Mat::zeros(vec_LTS[0].rows, vec_LTS[0].cols, vec_LTS[0].type());
  for (int i=0; i < dim; i++){
    accumulate (abs(vec_LTS[i]), sum_m, mask);
    accumulateSquare (vec_LTS[i], sum_sq, mask);
  }
  Scalar s_m = sum(sum_m);
  Scalar q_m = sum(sum_sq);
  stat_feat.at<double>(0, j++) = s_m.val[0]/sig_length;
  stat_feat.at<double>(0, j++) = q_m.val[0]/sig_length;
  
  //print_matrix (stat_feat, "Rows of stat_feat");      
  return (stat_feat);
}
double entropy (const Mat& data) {
  int i;
  
  double dE =0.0, d, eps = 0.00000000001;
  for (i=0; i<data.cols; i++) {
    d = data.at<double>(0,i);
    dE = dE - d*log(d+eps)-(1.0-d)*log(1.0-d+eps);
  }
  return dE;
}

int zero_crossing (const Mat& data){
  int i, j=0;
  int sign1, sign2;

  for(i=0; i<data.cols-1; i++) {
    sign1 = (data.at<double>(0, i)>0)?1:0;
    sign2 = (data.at<double>(0, i+1)>0)?1:0;
    if(sign1!=sign2) 
      j++;
  }
  return j;
}
double corr(const Mat& x, const Mat &y) {
  double var_x, var_y, s=0, s_x=0, s_y=0;
  Scalar mu_x, mu_y, sigma_x, sigma_y;
  MatND mask;

  meanStdDev (x, mu_x, sigma_x, mask);
  meanStdDev (y, mu_y, sigma_y, mask);

  for (int i=0; i< x.cols; i++) { 
    s = s+(x.at<double>(0,i) - mu_x.val[0])*(y.at<double>(0,i) - mu_y.val[0]);
    s_x = s_x + pow(x.at<double>(0,i) - mu_x.val[0], 2);
    s_y = s_y + pow(y.at<double>(0,i) - mu_y.val[0], 2);
  }
  return (double)s/(sqrt(s_x)*sqrt(s_y)); 
}
/******************************************************************/
/* END Program specific Core routines                           */
/******************************************************************/

} /* end of namespace */

int main(int argc, char** argv)
{
  P_PARAMS p;
  if (parse_options (argc, argv, p)==0){
    return process(p);
  } 
}
