#!/bin/bash
# Script to get shot descriptors for videos based on pre-computed 
# shot boundaries. 

# Input: <videofile.mp4> [example is provided]
#        <path/to/search/for/shot/file>
#        <path/to/output/directory>

# Output: <shotdesc_startframe_endframe.lie> [individual lie algebra representations for shot 1]
#         ... 
#         <shotdesc_startframe_endframe.lie> [individual lie algebra representations for shot N]
#         <videodesc.stat> [Statistical features computed on N shots, later stacked into Nxd matrix]

# 1. Given a video file, find the equivalent shot file.
# 2. Dump all frames from the video
# 3. Separate all frames into directories identified by shots frame numbers 
# 4. Create video files according to the shots
# 5. Run cameramotion descriptor code on each shot

if [ $# -ne 3 ]; then 
  echo "Usage: $0 <videofile.mp4> <path/to/search/for/shot/file> <output/dir>" 
  exit;
fi

video_name=`basename $1|sed 's/\.mp4//g'`;
shot_file=$video_name"_shots_v2.xml";
s_path=$2;
o_path="$3/$video_name";

if [ ! -s $1 ]; then
  echo "Input video file does not exist. Aborting."
  exit;
fi 

mkdir -p $o_path;

if [ ! -s $s_path/$shot_file ]; then
  echo "Shot file does not exist. Using the full video.";
  ./cameraMotionDesc -i $1 -w 4 -s $o_path/$video_name;
  exit;
fi

# Dump all frames from the video
echo "Dumping all frames from video $video_name";
FF_CMD="ffmpeg -y -loglevel quiet -i $1 -sameq $o_path/%05d.jpg >/dev/null 2>&1";
eval $FF_CMD;

# Parse shot file and create relevant directories
lines=`grep "<shot " $s_path/$shot_file|tr -s ' '|sed -e "s/<shot start_frame=\"//g"\
 -e "s/\" end_frame=\"/_/g" -e "s/\">//g" -e "s/\s//g"`;
for line in  $lines; do
  let st=`echo $line|cut -d_  -f1`;
  let ed=`echo $line|cut -d_  -f2`;
  shot_dir=$3/`printf "%s_%05d_%05d" $video_name $st $ed`;
  mkdir -p $shot_dir;
  echo "Moving relevant frames to $shot_dir";
  # ffmpeg expects frames to start with %01.jpg, hence renaming
  count=1;
  let st=$st+1;
  for file in `seq $st $ed`;do
    s_name=`printf "%05d.jpg" $file`;
    t_name=`printf "%05d.jpg" $count`;
    mv $o_path/$s_name $shot_dir/$t_name;
    let count=$count+1;
  done
  # Create shot videos
  echo "Creating video $shot_dir.mp4"
  ffmpeg -y -loglevel quiet -sameq -i $shot_dir/%05d.jpg $shot_dir.mp4 >/dev/null 2>&1; 
done

# Cleaning up intermediate files
find $3/ -mindepth 1 -type d -exec rm -rf {} \;
# Execute shot descriptor for each shot ID
for shot_file in `find $3 -name $video_name"_*.mp4"`; do 
  fname=`basename $shot_file|sed 's/\.mp4//g'`;
  ./cameraMotionDesc -i $shot_file -w 4 -s $3/$fname;
  cat $3/$fname.stat >> $3/$video_name.gstat;
  printf "\n" >> $3/$video_name.gstat;
done
# Cleaning up other files
find $3/ -name $video_name"_*.mp4" -exec rm -rf {} \;
find $3 -name $video_name"_*.stat" -exec rm -rf {} \; 
